open Processrc

type 'a par = Par of 'a array

let bsp_p_ref = ref 4
and bsp_g_ref = ref 1.0
and bsp_l_ref = ref 1.0
and bsp_time_start = ref 0.

let bsp_p () = !bsp_p_ref
let bsp_g () = !bsp_g_ref
let bsp_l () = !bsp_l_ref

let mkpar f = 
  Par(Array.init (bsp_p()) f)

let get (Par vv) (Par vi) =
  Par(Array.init (bsp_p()) (fun j->vv.(vi.(j))))

let put (Par (vf:(int->'a option) array)) = 
  Par(Array.init (bsp_p()) (fun i -> fun j -> (vf.(j)) i))

let apply (Par vf) (Par vv) =
  Par (Array.init (bsp_p()) (fun i-> (vf.(i)) (vv.(i))))

exception At_failure of string

let at (Par vb) i = 
 if not ((0<=i)&&(i<bsp_p()))
 then raise(At_failure "process not within bounds")
 else vb.(i)

let unsafe_proj (Par vv) = vv.(0)

exception Timer_failure of string

type timing_state = Running | Stopped

let timing = ref Stopped

let start_timing () = 
  if !timing=Stopped
  then 
    (bsp_time_start := Sys.time();
     timing := Running)
  else
    raise (Timer_failure "Timer is already running")

let stop_timing () =
  if !timing=Running
  then 
    (bsp_time_start := (Sys.time()) -. (!bsp_time_start);
     timing:=Stopped)
  else
   raise (Timer_failure "Timer was not started!")

let get_cost () = 
  mkpar(fun pid -> !bsp_time_start)

exception Error of string

let abort i s = 
  print_string s;
  flush stdout;
  exit i

let initialize () = 
  let params = readbsmllibrc None in
    begin
      (bsp_p_ref:=params.p;
       bsp_g_ref:=params.g;
       bsp_l_ref:=params.l);
    end

